import psutil
import os
import signal
import stat
import sys
import time
import traceback
from tempfile import NamedTemporaryFile
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
from shortuuid import ShortUUID
import atexit
from .config import Config
import logging
import threading
from six.moves import queue
import socket
import click
import wandb
from wandb import __stage_dir__, Error
from wandb import file_pusher
from wandb import sparkline
from wandb import stats
from wandb import streaming_log
from wandb import util
from .api import BinaryFilePolicy, CRDedupeFilePolicy
from .wandb_run import Run
logger = logging.getLogger(__name__)

ERROR_STRING = click.style('ERROR', bg='red', fg='green')


def editor(content='', marker='# Before we start this run, enter a brief description. (to skip, direct stdin to dev/null: `python train.py < /dev/null`)\n'):
    message = click.edit(content + '\n\n' + marker)
    if message is None:
        return None
    return message.split(marker, 1)[0].rstrip('\n')


class OutStreamTee(object):
    """Tees a writable filelike object.

    writes/flushes to the passed in stream will go to the stream
    and a second stream.
    """

    def __init__(self, stream, second_stream):
        """Constructor.

        Args:
            stream: stream to tee.
            second_stream: stream to duplicate writes to.
        """
        self._orig_stream = stream
        self._second_stream = second_stream
        self._queue = queue.Queue()
        self._thread = threading.Thread(target=self._thread_body)
        self._thread.daemon = True
        self._thread.start()

    def _thread_body(self):
        while True:
            item = self._queue.get()
            if item is None:
                break
            self._second_stream.write(item)

    def write(self, message):
        self._orig_stream.write(message)
        self._queue.put(message)

    def flush(self):
        self._orig_stream.flush()

    def close(self):
        self._queue.put(None)


class ExitHooks(object):
    def __init__(self):
        self.exit_code = None
        self.exception = None

    def hook(self):
        self._orig_exit = sys.exit
        self._orig_excepthook = sys.excepthook
        sys.exit = self.exit
        sys.excepthook = self.excepthook

    def exit(self, code=0):
        self.exit_code = code
        self._orig_exit(code)

    def excepthook(self, exc_type, exc, *args):
        self.exception = exc
        self._orig_excepthook(exc_type, exc, *args)


class FileTailer(object):
    def __init__(self, path, on_read_fn, binary=False):
        self._path = path
        mode = 'r'
        if binary:
            mode = 'rb'
        self._file = open(path, mode)
        self._on_read_fn = on_read_fn
        self._thread = threading.Thread(target=self._thread_body)
        self._thread.start()

    def _thread_body(self):
        while True:
            where = self._file.tell()
            data = self._file.read(1024)
            if not data:
                time.sleep(1)
                # required for to get python2 working (Issue #50)
                self._file.seek(where)
            else:
                self._on_read_fn(data)


class FileEventHandler(object):
    def __init__(self, file_path, save_name, api):
        self.file_path = file_path
        self.save_name = save_name
        self._api = api

    def on_created(self):
        pass

    def on_modified(self):
        pass

    def finish(self):
        pass


class FileEventHandlerOverwrite(FileEventHandler):
    def __init__(self, file_path, save_name, api, file_pusher, *args, **kwargs):
        super(FileEventHandlerOverwrite, self).__init__(
            file_path, save_name, api, *args, **kwargs)
        self._file_pusher = file_pusher

    def on_created(self):
        self.on_modified()

    def on_modified(self):
        # Tell file_pusher to copy the file, we want to allow the user to modify the
        # original while this one is uploading (modifying while uploading seems to
        # cause a hang somewhere in the google upload code, until the server times out)
        self._file_pusher.file_changed(
            self.save_name, self.file_path, copy=True)


class FileEventHandlerOverwriteDeferred(FileEventHandler):
    def __init__(self, file_path, save_name, api, file_pusher, *args, **kwargs):
        super(FileEventHandlerOverwriteDeferred, self).__init__(
            file_path, save_name, api, *args, **kwargs)
        self._file_pusher = file_pusher

    def finish(self):
        self._file_pusher.file_changed(self.save_name, self.file_path)


class FileEventHandlerTextStream(FileEventHandler):
    def __init__(self, *args, **kwargs):
        super(FileEventHandlerTextStream, self).__init__(*args, **kwargs)
        self._tailer = None

    def on_created(self):
        if self._tailer:
            logger.error(
                'Streaming file created twice in same run: %s', self.file_path)
            return
        self._setup()

    def on_modified(self):
        if self._tailer:
            return
        self._setup()

    def _setup(self):
        fsapi = self._api.get_file_stream_api()
        pusher = streaming_log.TextStreamPusher(fsapi, self.save_name)

        def on_read(data):
            pusher.write(data)

        self._tailer = FileTailer(self.file_path, on_read)


class FileEventHandlerBinaryStream(FileEventHandler):
    def __init__(self, *args, **kwargs):
        super(FileEventHandlerBinaryStream, self).__init__(*args, **kwargs)
        self._tailer = None

    def on_created(self):
        if self._tailer:
            logger.error(
                'Streaming file created twice in same run: %s', self.file_path)
            return
        self._setup()

    def on_modified(self):
        if self._tailer:
            return
        self._setup()

    def _setup(self):
        fsapi = self._api.get_file_stream_api()

        def on_read(data):
            fsapi.push(self.save_name, data)

        self._tailer = FileTailer(self.file_path, on_read, binary=True)


class Sync(object):
    """Watches for files to change and automatically pushes them
    """

    def __init__(self, api, run, config=None, project=None, tags=[], datasets=[], description=None):
        # 1.6 million 6 character combinations
        self._run = run
        self._watch_dir = os.path.abspath(self._run.dir)
        self._project = project or api.settings("project")
        self._entity = api.settings("entity")
        logger.debug("Initialized sync for %s/%s", self._project, self._run.id)

        # Load description and write it to the run directory.
        dpath = os.path.join(self._watch_dir, 'description.md')
        self._description = description
        if not self._description:
            if os.path.exists(dpath):
                with open(dpath) as f:
                    self._description = f.read()
            else:
                self._description = os.getenv('WANDB_DESCRIPTION')
        try:
            self.tty = (sys.stdin.isatty() and
                        os.getpgrp() == os.tcgetpgrp(sys.stdout.fileno()))  # check if background process
        except AttributeError:  # windows
            self.tty = sys.stdin.isatty()  # TODO Check for background process in windows

        if not self._description and self.tty:
            self._description = editor()
        if self._description is None:
            sys.stderr.write('No description provided, aborting run.\n')
            sys.exit(1)
        with open(dpath, 'w') as f:
            f.write(self._description)

        self._proc = psutil.Process(os.getpid())
        self._api = api
        self._tags = tags
        self._handler = PatternMatchingEventHandler()
        self._handler.on_created = self.on_file_created
        self._handler.on_modified = self.on_file_modified
        base_url = api.settings('base_url')
        if base_url.endswith('.dev'):
            base_url = 'http://app.dev'
        elif base_url.endswith('wandb.ai'):
            base_url = 'https://app.wandb.ai'
        self.url = "{base}/{entity}/{project}/runs/{run}".format(
            project=self._project,
            entity=self._entity,
            run=self._run.id,
            base=base_url
        )
        self._hooks = ExitHooks()
        self._hooks.hook()
        self._observer = Observer()

        self._observer.schedule(self._handler, self._watch_dir, recursive=True)

        if config is None:
            config = Config()
        self._config = config

        self._stats = stats.Stats()

        def push_function(save_name, path):
            with open(path, 'rb') as f:
                self._api.push(self._project, {save_name: f}, run=self._run.id,
                               progress=lambda _, total: self._stats.update_progress(path, total))
        self._file_pusher = file_pusher.FilePusher(push_function)

        self._event_handlers = {}
        # create a handler for description.md, so that we'll save it at the end
        # of the run.
        self._get_handler(dpath, 'description.md')

        try:
            signal.signal(signal.SIGQUIT, self._debugger)
        except AttributeError:
            pass

    def _debugger(self, *args):
        import pdb
        pdb.set_trace()

    def watch(self, files, show_run=False):
        try:
            # TODO: better failure handling
            upsert_result = self._api.upsert_run(name=self._run.id, project=self._project, entity=self._entity,
                                                 config=self._config.as_dict(), description=self._description, host=socket.gethostname())
            self._run_storage_id = upsert_result['id']
            self._handler._patterns = [
                os.path.join(self._watch_dir, os.path.normpath(f)) for f in files]
            # Ignore hidden files/folders
            self._handler._ignore_patterns = ['*/.*', '*.tmp']
            self._observer.start()

            self._api.save_patch(self._watch_dir)

            wandb.termlog("Syncing %s" % self.url)
            if show_run:
                import webbrowser
                webbrowser.open_new_tab(self.url)

            self._api.get_file_stream_api().set_file_policy(
                'output.log', CRDedupeFilePolicy())
            # Tee stdout/stderr into our TextOutputStream, which will push lines to the cloud.
            self._stdout_stream = streaming_log.TextStreamPusher(
                self._api.get_file_stream_api(), 'output.log', prepend_timestamp=True)
            sys.stdout = OutStreamTee(sys.stdout, self._stdout_stream)
            self._stderr_stream = streaming_log.TextStreamPusher(
                self._api.get_file_stream_api(), 'output.log', line_prepend='ERROR',
                prepend_timestamp=True)
            sys.stderr = OutStreamTee(sys.stderr, self._stderr_stream)

            self._stdout_stream.write(" ".join(psutil.Process(
                os.getpid()).cmdline()) + "\n\n")

            logger.debug("Swapped stdout/stderr")

            atexit.register(self.stop)
        except KeyboardInterrupt:
            self.stop()
        except Error:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            wandb.termlog('%s: An Exception was raised during setup, see %s for full traceback.' % (
                (ERROR_STRING, util.get_log_file_path())))
            wandb.termlog("%s: %s" % (ERROR_STRING, exc_value))
            lines = traceback.format_exception(
                exc_type, exc_value, exc_traceback)
            logger.error('\n'.join(lines))
            sys.exit(1)

    def update_config(self, config):
        self._config = config
        self._api.upsert_run(id=self._run_storage_id,
                             config=self._config.as_dict())

    def stop(self):
        wandb.termlog()
        if isinstance(self._hooks.exception, KeyboardInterrupt):
            wandb.termlog(
                'Script ended because of ctrl-c, press ctrl-c again to abort syncing.')
        elif self._hooks.exception:
            wandb.termlog(
                'Script ended because of Exception, press ctrl-c again to abort syncing.')
        else:
            wandb.termlog('Script ended.')

        # Show run summary/history
        summary = self._run.summary.summary
        if summary:
            wandb.termlog('Run summary:')
            max_len = max([len(k) for k in summary.keys()])
            format_str = '  {:>%s} {}' % max_len
            for k, v in summary.items():
                wandb.termlog(format_str.format(k, v))
        history_keys = self._run.history.keys()
        if history_keys:
            wandb.termlog('Run history:')
            max_len = max([len(k) for k in history_keys])
            for key in history_keys:
                vals = util.downsample(self._run.history.column(key), 40)
                line = sparkline.sparkify(vals)
                format_str = u'  {:>%s} {}' % max_len
                wandb.termlog(format_str.format(key, line))

        wandb.termlog('Waiting for final file modifications.')
        # This is a a heuristic delay to catch files that were written just before
        # the end of the script.
        # TODO: ensure we catch all saved files.
        time.sleep(2)
        try:
            self._observer.stop()
            self._observer.join()
        # TODO: py2 TypeError: PyCObject_AsVoidPtr called with null pointer
        except TypeError:
            pass
        # TODO: py3 SystemError: <built-in function stop> returned a result with an error set
        except SystemError:
            pass

        self._stdout_stream.close()
        self._stderr_stream.close()
        self._api.get_file_stream_api().finish(self._hooks.exception)

        for handler in self._event_handlers.values():
            handler.finish()
        self._file_pusher.finish()

        wandb.termlog('Syncing files in %s:' %
                      os.path.relpath(self._watch_dir))
        for file_path in self._stats.files():
            wandb.termlog('  %s' % os.path.relpath(file_path, self._watch_dir))
        step = 0
        spinner_states = ['-', '/', '|', '\\']
        stop = False
        self._stats.update_all_files()
        while True:
            if not self._file_pusher.is_alive():
                stop = True
            summary = self._stats.summary()
            line = (' %(completed_files)s of %(total_files)s files,'
                    ' %(uploaded_bytes).03f of %(total_bytes).03f bytes uploaded\r' % summary)
            line = spinner_states[step % 4] + line
            step += 1
            wandb.termlog(line, newline=False)
            if stop:
                break
            time.sleep(0.25)
            #print('FP: ', self._file_pusher._pending, self._file_pusher._jobs)
        # clear progress line.
        wandb.termlog(' ' * 79)

        # Check md5s of uploaded files against what's on the file system.
        # TODO: We're currently using the list of uploaded files as our source
        #     of truth, but really we should use the files on the filesystem
        #     (ie if we missed a file this wouldn't catch it).
        # This polls the server, because there a delay between when the file
        # is done uploading, and when the datastore gets updated with new
        # metadata via pubsub.
        wandb.termlog('Verifying uploaded files... ', newline=False)
        error = False
        mismatched = None
        for delay_base in range(4):
            mismatched = []
            download_urls = self._api.download_urls(
                self._project, run=self._run.id)
            for fname, info in download_urls.items():
                if fname == 'wandb-history.h5' or 'output.log':
                    continue
                local_path = os.path.join(self._watch_dir, fname)
                local_md5 = util.md5_file(local_path)
                if local_md5 != info['md5']:
                    mismatched.append((local_path, local_md5, info['md5']))
            if not mismatched:
                break
            wandb.termlog('  Retrying after %ss' % (delay_base**2))
            time.sleep(delay_base ** 2)

        if mismatched:
            print('')
            error = True
            for local_path, local_md5, remote_md5 in mismatched:
                wandb.termlog(
                    '%s: %s (%s) did not match uploaded file (%s) md5' % (
                        ERROR_STRING, local_path, local_md5, remote_md5))
        else:
            print('verified!')

        if error:
            wandb.termlog('%s: Sync failed %s' % (ERROR_STRING, self.url))
        else:
            wandb.termlog('Synced %s' % self.url)

    def _get_handler(self, file_path, save_name):
        self._stats.update_file(file_path)
        if save_name not in self._event_handlers:
            if save_name == 'wandb-history.jsonl':
                self._event_handlers['wandb-history.jsonl'] = FileEventHandlerTextStream(
                    file_path, 'wandb-history.jsonl', self._api)
            # Don't try to stream tensorboard files for now.
            # elif 'tfevents' in save_name:
            #    # TODO: This is hard-coded, but we want to give users control
            #    # over streaming files (or detect them).
            #    self._api.get_file_stream_api().set_file_policy(save_name,
            #                                                    BinaryFilePolicy())
            #    self._event_handlers[save_name] = FileEventHandlerBinaryStream(
            #        file_path, save_name, self._api)
            # Overwrite handler (non-deferred) has a bug, wherein if the file is truncated
            # during upload, the request to Google hangs (at least, this is my working
            # theory). So for now we defer uploading everything til the end of the run.
            # TODO: send wandb-summary during run. One option is to copy to a temporary
            # file before uploading.
            elif save_name == 'wandb-summary.json':
                self._event_handlers[save_name] = FileEventHandlerOverwrite(
                    file_path, save_name, self._api, self._file_pusher)
            else:
                self._event_handlers[save_name] = FileEventHandlerOverwriteDeferred(
                    file_path, save_name, self._api, self._file_pusher)
        return self._event_handlers[save_name]

    # TODO: limit / throttle the number of adds / pushes
    def on_file_created(self, event):
        if os.path.isdir(event.src_path):
            return None
        save_name = os.path.relpath(event.src_path, self._watch_dir)
        self._get_handler(event.src_path, save_name).on_created()

    # TODO: is this blocking the main thread?
    def on_file_modified(self, event):
        if os.path.isdir(event.src_path):
            return None
        save_name = os.path.relpath(event.src_path, self._watch_dir)
        self._get_handler(event.src_path, save_name).on_modified()

    @property
    def source_proc(self):
        mode = os.fstat(0).st_mode
        if not stat.S_ISFIFO(mode):
            # stdin is not a pipe
            return None
        else:
            source = self._proc.parent().children()[0]
            return None if source == self._proc else source



import pytest
import os
from click.testing import CliRunner
import git
from wandb.git_repo import GitRepo


@pytest.fixture
def git_repo():
    with CliRunner().isolated_filesystem():
        r = git.Repo.init(".")
        open("README", "wb").close()
        r.index.add(["README"])
        r.index.commit("Initial commit")
        yield GitRepo(lazy=False)



import yaml
import os
import sys
import logging
from .api import Error

logger = logging.getLogger(__name__)


def boolify(s):
    if s.lower() == 'none':
        return None
    if s.lower() == 'true':
        return True
    if s.lower() == 'false':
        return False
    raise ValueError("Not a boolean")


class Config(object):
    """Creates a W&B config object."""

    def __init__(self, config_paths=[], wandb_dir=None, run_dir=None, persist_callback=None):
        object.__setattr__(self, '_wandb_dir', wandb_dir)
        object.__setattr__(self, '_run_dir', run_dir)

        # TODO: Replace this with an event system.
        object.__setattr__(self, '_persist_callback', persist_callback)

        object.__setattr__(self, '_items', {})
        object.__setattr__(self, '_descriptions', {})

        self._load_defaults()
        for conf_path in config_paths:
            self._load_file(conf_path)
        self._persist()

    def _load_defaults(self):
        if not self._wandb_dir:
            logger.debug('wandb dir not provided, skipping defaults')
            return
        defaults_path = os.path.join('config-defaults.yaml')
        if not os.path.exists(defaults_path):
            logger.debug('no defaults not found in %s' % defaults_path)
            return
        self._load_file(defaults_path)

    def _load_file(self, path):
        subkey = None
        if '::' in path:
            conf_path, subkey = path.split('::', 1)
        else:
            conf_path = path
        try:
            conf_file = open(conf_path)
        except (OSError, IOError):
            raise Error('Couldn\'t read config file: %s' % conf_path)
        try:
            loaded = yaml.load(conf_file)
        except yaml.parser.ParserError:
            raise Error('Invalid YAML in config-defaults.yaml')
        if subkey:
            try:
                loaded = loaded[subkey]
            except KeyError:
                raise Error('Asked for %s but %s not present in %s' % (
                    path, subkey, conf_path))
        self._update_from_dict(
            loaded, error_prefix='In config %s, ' % path)

    def _update_from_dict(self, items, error_prefix=''):
        if not isinstance(items, dict):
            raise Error('%sExpected dict but received %s' %
                        (error_prefix, items))
        for key, val in items.items():
            if key == 'wandb_version':
                continue
            self._set_key(key, val, error_prefix=error_prefix)

    def _set_key(self, key, val, error_prefix=''):
        if isinstance(val, dict):
            if 'value' not in val:
                raise Error('%svalue of %s is dict, but does not contain "value" key' % (
                    error_prefix, key))
            self._items[key] = val['value']
            if 'desc' in val:
                self._descriptions[key] = val['desc']
        else:
            self._items[key] = val

    def keys(self):
        """All keys in the current configuration"""
        return self._items.keys()

    def desc(self, key):
        """The description of a given key"""
        return self._descriptions.get(key)

    def load_json(self, json):
        """Loads existing config from JSON"""
        for key in json:
            self._items[key] = json[key].get('value')
            self._descriptions[key] = json[key].get('desc')

    def _persist(self):
        """Stores the current configuration for pushing to W&B"""
        if not self._run_dir or not os.path.isdir(self._run_dir):
            # In dryrun mode, without wandb run, we don't
            # save config  on initial load, because the run directory
            # may not be created yet (because we don't know if we're
            # being used in a run context, or as an API).
            # TODO: Defer saving somehow, maybe via an events system
            return
        path = os.path.join(self._run_dir, 'config.yaml')
        with open(path, "w") as conf_file:
            conf_file.write(str(self))

        # TODO: Replace with events
        if self._persist_callback:
            self._persist_callback()

    def get(self, *args):
        return self._items.get(*args)

    def __getitem__(self, key):
        return self._items[key]

    def __setitem__(self, key, val):
        self._set_key(key, val)
        self._persist()

    __setattr__ = __setitem__

    def __getattr__(self, key):
        return self.__getitem__(key)

    def update(self, params):
        if not isinstance(params, dict):
            try:
                # for tensorflow flags
                if "__flags" in dir(params):
                    if not params.__dict__['__parsed']:
                        params._parse_flags()
                    params = params.__dict__['__flags']
                else:
                    params = vars(params)
            except TypeError:
                raise TypeError(
                    "config must be a dict or have a __dict__ attribute.")
        self._update_from_dict(params)
        self._persist()

    def as_dict(self):
        defaults = {}
        for key, val in self._items.items():
            defaults[key] = {'value': val,
                             'desc': self._descriptions.get(key)}
        return defaults

    def __str__(self):
        s = "wandb_version: 1\n\n"
        s += yaml.dump(self.as_dict(), default_flow_style=False)
        return s



import datetime
import os
import shortuuid

import wandb
from wandb import history
from wandb import summary
from wandb import util


class Run(object):
    def __init__(self, run_id, dir, config):
        self.id = run_id
        self._dir = os.path.abspath(dir)
        self.config = config
        self._made_dir = False

        # Use internal self._dir so dir isn't created yet (it should only
        # be created when the user accesses a run member for the first time).
        self._history = None
        self._summary = None

    def _mkdir(self):
        if self._made_dir:
            return
        util.mkdir_exists_ok(self._dir)

    @property
    def dir(self):
        self._mkdir()
        return self._dir

    @property
    def summary(self):
        self._mkdir()
        if self._summary is None:
            self._summary = summary.Summary(self._dir)
        return self._summary

    @property
    def history(self):
        self._mkdir()
        if self._history is None:
            self._history = history.History(self._dir)
        return self._history


def generate_id():
    # ~3t run ids (36**8)
    run_gen = shortuuid.ShortUUID(alphabet=list(
        "0123456789abcdefghijklmnopqrstuvwxyz"))
    return run_gen.random(8)


def run_dir_path(run_id, dry=False):
    if dry:
        prefix = 'dryrun'
    else:
        prefix = 'run'
    time_str = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
    return os.path.join(wandb.wandb_dir(), '%s-%s-%s' % (prefix, time_str, run_id))



import datetime
import io
import tempfile
import time
import sys
from requests import Session
from six.moves import queue
import traceback
import logging
import re
import signal
import os
import logging
import six
logger = logging.getLogger(__name__)


class LineBuffer(object):
    """Streaming string parser that extracts lines."""

    def __init__(self):
        self._buf = []
        self._line_end_re = re.compile('[\r\n]')

    def add_string(self, string):
        """Process a string.

        Args:
            string: Any string
        Returns:
            list of found lines, remainder will be buffered.
        """
        lines = []
        while string:
            match = self._line_end_re.search(string)
            if match is None:
                self._buf.append(string)
                break
            else:
                line_end_pos = match.start()
                lines.append(''.join(self._buf) + string[:line_end_pos + 1])
                string = string[line_end_pos + 1:]
                self._buf = []
        return lines


class TextStreamPusher(object):
    """Pushes a stream of text, line by line, to wandb."""

    def __init__(self, fsapi, filename, line_prepend='', prepend_timestamp=False):
        """Conctructor.

        Args:
            fsapi: api.FileStreamApi instance
            filename: Name of the file this stream is pushed to.
            line_prepend: string to prepend to every line for this stream.
            prepend_timestamp: If true a timestamp will be prepended to each line
                (after line_prepend).
        """
        self._fsapi = fsapi
        self._filename = filename
        if line_prepend:
            line_prepend += ' '
        self._line_prepend = line_prepend
        self._prepend_timestamp = prepend_timestamp
        self._line_buffer = LineBuffer()

    def write(self, message, cur_time=None):
        """Write some text to the pusher.

        Args:
            message: a string to push for this file.
            cur_time: used for unit testing. override line timestamp.
        """
        if cur_time is None:
            cur_time = time.time()
        lines = self._line_buffer.add_string(message)
        for line in lines:
            timestamp = ''
            if self._prepend_timestamp:
                timestamp = datetime.datetime.utcfromtimestamp(
                    cur_time).isoformat() + ' '
            line = '%s%s%s' % (self._line_prepend, timestamp, line)
            self._fsapi.push(self._filename, line)

    def close(self):
        """Close the file."""
        # Force a final line to clear whatever might be in the buffer.
        self.write('\n')



import csv
import os

import wandb
from wandb import util

HISTORY_FNAME = 'wandb-history.jsonl'


class History(object):
    """Used to store data that changes over time during runs. """

    def __init__(self, out_dir='.'):
        self.fname = os.path.join(out_dir, HISTORY_FNAME)
        self._file = open(self.fname, 'w')
        self.rows = []

    def keys(self):
        if self.rows:
            return self.rows[0].keys()
        return []

    def column(self, key):
        return [r[key] for r in self.rows]

    def add(self, row):
        if not isinstance(row, dict):
            raise wandb.Error('history.add expects dict')
        self.rows.append(row)
        self._file.write(util.json_dumps_safer(row))
        self._file.write('\n')
        self._file.flush()



import collections
import os

class FileStats(object):
    def __init__(self, file_path):
        self._file_path = file_path
        self.size = 0
        self.uploaded = 0
    
    def update_size(self):
        self.size = os.path.getsize(self._file_path)


class Stats(object):
    def __init__(self):
        self._files = {}

    def update_file(self, file_path):
        if file_path not in self._files:
            self._files[file_path] = FileStats(file_path)
        self._files[file_path].update_size()

    def update_all_files(self):
        for file_stats in self._files.values():
            file_stats.update_size()

    def update_progress(self, file_path, uploaded):
        if file_path in self._files:
            self._files[file_path].uploaded = uploaded

    def files(self):
        return self._files.keys()

    def stats(self):
        return self._files

    def summary(self):
        return {
            'completed_files': sum(f.size == f.uploaded for f in self._files.values()),
            'total_files': len(self._files),
            'uploaded_bytes': sum(f.uploaded for f in self._files.values()),
            'total_bytes': sum(f.size for f in self._files.values())
        }


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.md') as readme_file:
    readme = readme_file.read()

requirements = [
    'Click>=6.0',
    'gql>=0.1.0',
    'requests>=2.0.0',
    'whaaaaat>=0.5.2',
    'six>=1.10.0',
    'psutil>=5.2.2',
    'watchdog>=0.8.3',
    'GitPython>=2.1.3',
    'shortuuid>=0.5.0'
]

test_requirements = [
    'mock>=2.0.0',
    'tox-pyenv>=1.0.3'
]

setup(
    name='wandb',
    version='0.4.28',
    description="A CLI and library for interacting with the Weights and Biases API.",
    long_description=readme,
    author="Chris Van Pelt",
    author_email='vanpelt@gmail.co',
    url='https://github.com/wandb/client',
    packages=[
        'wandb',
    ],
    package_dir={'wandb':
                 'wandb'},
    entry_points={
        'console_scripts': [
            'wandb=wandb.cli:cli',
            'wb=wandb.cli:cli',
            'wanbd=wandb.cli:cli'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='wandb',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



# -*- coding: utf-8 -*-

__author__ = """Chris Van Pelt"""
__email__ = 'vanpelt@wandb.com'
__version__ = '0.4.28'

import click
import types
import six
import sys
import logging
import os
import traceback

# We use the hidden version if it already exists, otherwise non-hidden.
if os.path.exists('.wandb'):
    __stage_dir__ = '.wandb/'
elif os.path.exists('wandb'):
    __stage_dir__ = "wandb/"
else:
    __stage_dir__ = None


def wandb_dir():
    return __stage_dir__


def _set_stage_dir(stage_dir):
    # Used when initing a new project with "wandb init"
    global __stage_dir__
    __stage_dir__ = stage_dir


if __stage_dir__ is not None:
    log_fname = __stage_dir__ + 'debug.log'
else:
    log_fname = './wandb-debug.log'
logging.basicConfig(
    filemode="w",
    filename=log_fname,
    level=logging.DEBUG)


class Error(Exception):
    """Base W&B Error"""
    # For python 2 support

    def encode(self, encoding):
        return self.message


# These imports need to be below __stage_dir__ declration until we remove
# 'from wandb import __stage_dir__' from api.py etc.
from wandb import api as wandb_api
from wandb import config as wandb_config
from wandb import sync
from wandb import wandb_run

# Three possible modes:
#     'cli': running from "wandb" command
#     'run': we're a script launched by "wandb run"
#     'dryrun': we're a script not launched by "wandb run"


_orig_stderr = sys.stderr


def termlog(string='', newline=True):
    if string:
        line = '%s: %s' % (click.style('wandb', fg='blue', bold=True), string)
    else:
        line = ''
    click.echo(line, file=_orig_stderr, nl=newline)


def _do_sync(mode, run, show_run, extra_config=None):
    syncer = None
    termlog()
    if mode == 'run':
        api = wandb_api.Api()
        if api.api_key is None:
            raise wandb_api.Error(
                "No API key found, run `wandb login` or set WANDB_API_KEY")
        api.set_current_run_id(run.id)
        if extra_config is not None:
            run.config.update(extra_config)
        syncer = sync.Sync(api, run, config=run.config)
        syncer.watch(files='*', show_run=show_run)
    elif mode == 'dryrun':
        termlog(
            'wandb dryrun mode. Use "wandb run <script>" to save results to wandb.')
    termlog('Run directory: %s' % os.path.relpath(run.dir))
    termlog()
    return syncer


# Will be set to the run object for the current run, as returned by
# wandb.init(). We may want to get rid of this, but WandbKerasCallback
# relies on it, and it improves the API a bit (user doesn't have to
# pass the run into WandbKerasCallback)
run = None


def init():
    # The WANDB_DEBUG check ensures tests still work.
    if not __stage_dir__ and not os.getenv('WANDB_DEBUG'):
        print('wandb.init() called but directory not initialized.\n'
              'Please run "wandb init" to get started')
        sys.exit(1)

    # parse environment variables
    mode = os.getenv('WANDB_MODE', 'dryrun')
    run_id = os.getenv('WANDB_RUN_ID')
    if run_id is None:
        run_id = wandb_run.generate_id()
    run_dir = os.getenv('WANDB_RUN_DIR')
    if run_dir is None:
        run_dir = wandb_run.run_dir_path(run_id, dry=mode == 'dryrun')
    conf_paths = os.getenv('WANDB_CONFIG_PATHS', '')
    if conf_paths:
        conf_paths = conf_paths.split(',')
    show_run = bool(os.getenv('WANDB_SHOW_RUN'))

    config = None
    syncer = None

    def persist_config_callback():
        if syncer:
            syncer.update_config(config)
    config = wandb_config.Config(config_paths=conf_paths,
                                 wandb_dir=__stage_dir__, run_dir=run_dir,
                                 persist_callback=persist_config_callback)
    global run
    run = wandb_run.Run(run_id, run_dir, config)
    syncer = _do_sync(mode, run, show_run)
    return run


__all__ = ['init', 'termlog', 'run']
